import { Widget, VDOM } from "../ui/Widget";
import { HtmlElement } from "../widgets/HtmlElement";
import { PureContainer } from "../ui/PureContainer";
import { getShape } from "./shapes";
import { isUndefined } from "../util/isUndefined";
import { isArray } from "../util/isArray";
import { withHoverSync } from "../ui/HoverSync";

export class Legend extends HtmlElement {
   prepareData(context, instance) {
      let { data } = instance;
      data.stateMods = Object.assign(data.stateMods || {}, {
         vertical: this.vertical,
      });
      super.prepareData(context, instance);
   }

   declareData() {
      super.declareData(...arguments, {
         shape: undefined,
      });
   }

   isValidHtmlAttribute(attrName) {
      switch (attrName) {
         case "shapeSize":
         case "svgSize":
            return false;

         default:
            return super.isValidHtmlAttribute(attrName);
      }
   }

   explore(context, instance) {
      if (!context.legends) context.legends = {};

      instance.legends = context.legends;

      context.addLegendEntry = (legendName, entry) => {
         //case when all legends are scoped and new entry is added outside the scope
         if (!context.legends) return;

         let legend = context.legends[legendName];
         if (!legend)
            legend = context.legends[legendName] = {
               entries: [],
               names: {},
            };

         if (!legend.names[entry.name]) {
            legend.entries.push(entry);
            legend.names[entry.name] = entry;
         }
      };

      super.explore(context, instance);
   }

   renderChildren(context, instance) {
      const CSS = this.CSS;

      let entries = instance.legends[this.name] && instance.legends[this.name].entries,
         list;

      if (isArray(entries) && entries.length > 0) {
         list = (
            <div key="wrap" className={CSS.element(this.baseClass, "wrap")}>
               {entries.map((e, i) =>
                  withHoverSync(i, e.hoverSync, e.hoverChannel, e.hoverId, ({ onMouseMove, onMouseLeave, hover }) => (
                     <div
                        key={i}
                        className={CSS.element(this.baseClass, "entry", {
                           "color-root": true,
                           hover,
                           disabled: e.disabled,
                           selected: e.selected,
                        })}
                        onClick={e.onClick}
                        onMouseMove={onMouseMove}
                        onMouseLeave={onMouseLeave}
                     >
                        {this.renderShape(e, instance.data.shape)}
                        {e.name}
                     </div>
                  ))
               )}
            </div>
         );
      }

      return [list, super.renderChildren(context, instance)];
   }

   renderShape(entry, legendEntriesShape) {
      const className = this.CSS.element(this.baseClass, "shape", {
         [`color-${entry.colorIndex}`]: entry.colorIndex != null && (isUndefined(entry.active) || entry.active),
      });
      const shape = getShape(legendEntriesShape || entry.shape || "square");

      return (
         <svg
            className={this.CSS.element(this.baseClass, "svg")}
            style={{
               width: `${this.svgSize}px`,
               height: `${this.svgSize}px`,
               marginTop: `${-this.svgSize / 2}px`,
            }}
         >
            {shape(this.svgSize / 2, this.svgSize / 2, entry.shapeSize || this.shapeSize, {
               style: entry.style,
               className: className,
            })}
         </svg>
      );
   }
}

Legend.prototype.name = "legend";
Legend.prototype.baseClass = "legend";
Legend.prototype.vertical = false;
Legend.prototype.memoize = false;
Legend.prototype.shapeSize = 18;
Legend.prototype.shape = null;
Legend.prototype.svgSize = 20;

Widget.alias("legend", Legend);

Legend.Scope = class extends PureContainer {
   explore(context, instance) {
      context.push("legends", (instance.legends = {}));
      super.explore(context, instance);
   }

   exploreCleanup(context, instance) {
      context.pop("legends");
   }

   prepare(context, instance) {
      context.push("legends", instance.legends);
   }

   prepareCleanup(context, instance) {
      context.pop("legends");
   }
};

export const LegendScope = Legend.Scope;
